/*
Copyright (c) 1997-2005, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.hd.d.tmf;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;


/**Counts the bytes it passes downstream.
 * It does not necessarily correctly account for discarded
 * output caused by (for example) a buffer reset before commit.
 * <p>
 * While we are blocked writing output we temporarily decrement
 * the connection count on the grounds that we are I/O bound
 * rather than internally resource bound.
 */
public final class ThroughputMonitorResponseStream extends ServletOutputStream
    {
    /**Construct a new filter to monitor output flow.
     * Package visible to allow construction from the ThroughputMonitorFilter.
     */
    ThroughputMonitorResponseStream(final ThroughputMonitorFilterBase f,
                                    final HttpServletResponse response)
        throws IOException
        {
        super();
        stream = response.getOutputStream();
        filter = f;
        }

    /**The original output stream... */
    private final ServletOutputStream stream;

    /**The throughput filter... */
    private final ThroughputMonitorFilterBase filter;

    /**Write a whole byte array to the output stream. */
    public void write(final byte[] b)
        throws IOException
        { write(b, 0, b.length); }

    /**Write a single byte to the output stream; very inefficient. */
    public void write(final int b)
        throws IOException
        {
        final boolean didDecrement = filter.adjustBusyConnectionCount(false);
        try { stream.write(b); }
        finally { if(didDecrement) { filter.adjustBusyConnectionCount(true); } }
        filter.moreBytes(1);
        }

    /**Write part of byte array to the output stream. */
    public void write(final byte[] b, final int off, final int len)
        throws IOException
        {
        final boolean didDecrement = filter.adjustBusyConnectionCount(false);
        try { stream.write(b, off, len); }
        finally { if(didDecrement) { filter.adjustBusyConnectionCount(true); } }
        filter.moreBytes(len);
        }

    /**
     * Flushes this output stream and forces any buffered output bytes
     * to be written out. The general contract of <code>flush</code> is
     * that calling it is an indication that, if any bytes previously
     * written have been buffered by the implementation of the output
     * stream, such bytes should immediately be written to their
     * intended destination.
     * <p>
     * The <code>flush</code> method of <code>OutputStream</code> does nothing.
     *
     * @exception  IOException  if an I/O error occurs.
     */
    public void flush()
        throws IOException
        {
        final boolean didDecrement = filter.adjustBusyConnectionCount(false);
        try { stream.flush(); }
        finally { if(didDecrement) { filter.adjustBusyConnectionCount(true); } }
        }

    /**
     * Closes this output stream and releases any system resources
     * associated with this stream. The general contract of <code>close</code>
     * is that it closes the output stream. A closed stream cannot perform
     * output operations and cannot be reopened.
     * <p>
     * The <code>close</code> method of <code>OutputStream</code> does nothing.
     *
     * @exception  IOException  if an I/O error occurs.
     */
    public void close() throws IOException
        {
        final boolean didDecrement = filter.adjustBusyConnectionCount(false);
        try { stream.close(); }
        finally { if(didDecrement) { filter.adjustBusyConnectionCount(true); } }
        }
    }
